@@ -29,6 +29,14 @@ class KestrelInternalError(KestrelException):
         super().__init__(error, "please open a github issue to report")
 
 
+class KestrelNotImplemented(KestrelException):
+    def __init__(self, error):
+        super().__init__(
+            "Functionality not implemented: " + error,
+            "please search for the github issue to add comment",
+        )
+
+
 ################################################################
 #                     Kestrel Session Errors
 ################################################################
@@ -91,6 +99,10 @@ class InvalidStixPattern(KestrelException):
         super().__init__(msg, "rewrite the STIX pattern")
 
 
+class InvalidECGPattern(KestrelException):
+    pass
+
+
 class VariableNotExist(KestrelException):
     def __init__(self, var_name):
         super().__init__(
